"""Module covering the test suite to make sure your catalog is set up properly.

.. include:: ../../gh_pages/test_suite.md
"""
from . import catalog_tests
from . import column_store_tests
from . import description_tests


__all__ = [
    'catalog_tests',
    'column_store_tests',
    'description_tests'
]



"""Module containing the Catalog class and helper functions."""
from __future__ import annotations
from typing import Optional, Sequence, Dict, Union
from collections import defaultdict
from hela._constants import _PATH_VAR
from hela.plots.html_displays import DFDisplay
from hela._utils.path_utils import join_paths
from hela.math import levenshtein, tf_idf
from hela._base_dataset import BaseDataset
from hela._column_classes import _ColumnType
from hela._utils.short_description import ShortDescription


def is_catalog(obj: object) -> Union[Catalog, None]:
    if getattr(obj, '_is_catalog', False):
        return obj
    return None


def is_dataset(obj: object) -> Union[BaseDataset, None]:
    if getattr(obj, '_is_dataset', False):
        return obj
    return None


class Catalog:
    """Inheritable Catalog class, used when building your own data catalog.

    The namesake of the python package, this class will turn your code from just being code into a data catalog.
    This class will make your datasets iterable, testable, referenceable and more.
    You can also decorate the catalog with the `Catalog.setup` function, giving your catalog a description and
    enriching datasets within it.

    Examples:
    >>> # Without decorator
    >>> from catalog import Catalog
    >>> class MyCatalog(Catalog):
    ...     my_dataset = Dataset(...)

    >>> # With decorator
    >>> from catalog import Catalog
    >>> @Catalog.setup(folder='sales', description='Datasets related to sales.')
    >>> class SalesCatalog(Catalog):
    ...     my_dales_dataset = Dataset(...)
    """
    _is_catalog: bool = True
    _type: str = 'Catalog'
    _folder: str = None
    _database: str = None
    _description: str = None
    _rich_description_path: str = None

    @staticmethod
    def setup(
        cls: Catalog = None,
        folder: Optional[str] = None,
        database: Optional[str] = None,
        description: Optional[str] = None,
        rich_description_path: Optional[str] = None
    ) -> Catalog:
        """Decorator enriching the catalog with a description, and optionally binding
        a folder or database to all datasets within it.

        Args:
            folder: Used for filestore style datasets (e.g. spark),
                    build the catalogs folder structure.
            database:   Used for database style datasets (e.g. bigquery,
                        aws glue) builds the catalogs database structure.
            description:    A description of this catalog.
            rich_description_path: Path to markdown file with richer descriptions of this catalog.
        """

        def wrap(cls: Catalog):
            if not getattr(cls, '_is_catalog', False):
                raise ValueError(f'Class {cls} must inherit Catalog class.')
            cls._folder = folder
            cls._database = database
            cls._description = description
            cls._rich_description_path = rich_description_path
            _enrich_datasets(cls, folder=folder, database=database)
            return cls

        # When called with pathentheses "@catalog()"
        if cls is None:
            return wrap

        # When called without pathentheses "@catalog"
        return wrap(cls)

    @classmethod
    def get_catalogs(cls, recursive: bool = True) -> Sequence[Catalog]:
        """Get a list of all sub-catalogs of this catalog, not including self.

        Args:
            recursive:  Flag whether to search for sub-catalog in this catalog's sub-catalogs.

        Returns:
            A list of objects inheriting the Catalog class.
        """
        catalog_list = []
        for obj in cls.__dict__.values():
            catalog = is_catalog(obj)
            if catalog:
                catalog_list.append(catalog)
                if recursive:
                    catalog_list.extend(catalog.get_catalogs())
        return catalog_list

    @classmethod
    def get_datasets(cls, recursive: bool = True) -> Sequence[BaseDataset]:
        """Returns a list of all datasets within this catalog, recursively if flag is set.

        Args:
            recursive: When set to true this function will fetch datasets from sub-catalogs of this catalog.

        Returns:
            A list of dataset objects.
        """
        dataset_list = []
        for obj in cls.__dict__.values():
            dataset = is_dataset(obj)
            if dataset:
                dataset_list.append(obj)
            elif recursive:
                catalog = is_catalog(obj)
                if catalog:
                    dataset_list.extend(catalog.get_datasets())
        return dataset_list

    @classmethod
    def get_columns_datasets(cls, recursive: bool = True) -> Dict[_ColumnType, Sequence[BaseDataset]]:
        column_dict = defaultdict(list)
        for dataset in cls.get_datasets(recursive=recursive):
            if dataset.columns is None:
                continue

            for column in dataset.columns:
                column_dict[column].append(dataset)
        return column_dict

    @classmethod
    def show_datasets(cls, recursive: bool = True) -> DFDisplay:
        """Returns a DFDisplay with a description of all datasets in this catalog, one dataset per row.

        Args:
            recursive: Whether to show dataset recursively in subcatalogs.

        Returns:
            DFDisplay (pandas dataframe)
        """
        return DFDisplay([
            d._describe().__dict__
            for d in cls.get_datasets(recursive=recursive)
        ])

    @classmethod
    def show_columns(cls, recursive: bool = True) -> DFDisplay:
        """Returns a pandas dataframe with a description of all columns in this catalog, one column per row.

        Args:
            recursive: Whether to show dataset recursively in subcatalogs.

        Returns:
            DFDisplay (pandas dataframe)
        """
        df = DFDisplay([
            {
                **col.__dict__,
                'datasets': datasets
            }
            for col_obj, datasets in cls.get_columns_datasets(recursive=recursive).items()
            for col in col_obj._describe()
        ])
        return df.sort_values('name').reset_index(drop=True)

    @classmethod
    def _all_descriptions(cls, recursive=True) -> Sequence[ShortDescription]:
        """Collect descriptions for all Catalogs, Datasets and Columns.

        Args:
            recursive:  Whether to search recursively through sub-catalogs.

        Returns:
            A sequence of ShortDescription objects
        """
        column_descriptions = []
        for c in cls.get_columns_datasets(recursive=recursive).keys():
            desc = c._desc_()
            # Columns will sometimes give a sequence if it is a nested column
            if isinstance(desc, Sequence):
                column_descriptions.extend(desc)
            else:
                column_descriptions.append(desc)
        return [
            *column_descriptions,
            *[d._desc_() for d in cls.get_datasets(recursive=recursive)],
            *[c._desc_() for c in cls.get_catalogs(recursive=recursive)],
            cls._desc_()
        ]

    @classmethod
    def search(cls, query_str: str, recursive: bool = True, max_hits: int = 5, min_relevance: float = .1) -> DFDisplay:
        """Searches across names and descriptions of Catalogs, Datasets and Columns.

        * Search on name is based on Levenshtein distance (fuzzy search).
        * Search on description is based on cosine similarity of TF-IDF matrix.

        Args:
            query_str:  The string to base the search on.
            recursive:  Whether to search recursively through sub-catalogs.
            max_hits:   Maximum number of hits to return.
            min_relevance: Minimum required relevance score to return a hit.

        Returns:
            A DFDisplay (pandas) dataframe with hits sorted on relevance.
        """

        # Create dictionaries to collect into unique name/desc keys
        name_dict, desc_dict = defaultdict(list), defaultdict(list)
        for short_desc in cls._all_descriptions(recursive=recursive):
            name = short_desc.name
            if name is not None:
                name_dict[name].append(short_desc)
                # Take both original and split version if we're dealing with nested columns
                # E.g. ratings.taste
                if '.' in name:
                    name_dict[name.split('.')[-1]].append(short_desc)
            if short_desc.description is not None:
                desc_dict[short_desc.description].append(short_desc)

        l_search = levenshtein.sort(query_str, list(name_dict.keys()), min_similarity=.5)
        try:
            tf_idf_search = tf_idf.sort(query_str, list(desc_dict.keys()))
        except ValueError:
            # Raised when we get no vocabulary matches at all
            tf_idf_search = []

        relevance_col = 'relevance'

        hit_df = DFDisplay([
            *[
                {
                    **short_desc.__dict__,
                    relevance_col: hit.score,
                    'hit_on': 'name'
                }
                for hit in l_search
                for short_desc in name_dict[hit.match_string]
            ],
            *[
                {
                    **short_desc.__dict__,
                    relevance_col: hit.score,
                    'hit_on': 'description'
                }
                for hit in tf_idf_search
                for short_desc in desc_dict[hit.match_string]
            ]
        ])
        error_msg = f'No hits good enough found on query string: "{query_str}"'
        if len(hit_df) == 0:
            raise ValueError(error_msg)
        hit_df = hit_df[hit_df[relevance_col] > min_relevance]

        if len(hit_df) == 0:
            raise ValueError(error_msg)

        hit_df.loc[:, relevance_col] = hit_df[relevance_col].round(2)
        return (
            hit_df
            .sort_values(relevance_col, ascending=False)
            .drop_duplicates(subset=['name', 'type'])
            [:max_hits]
        )

    @classmethod
    def _desc_(self) -> ShortDescription:
        return ShortDescription(name=self.__name__, type=self._type, description=self._description)


def _enrich_datasets(
        catalog: Catalog,
        folder: Optional[str] = None,
        database: Optional[str] = None) -> None:
    """Perform various enrichments such as extending with path, database and adding its own catalog.

    Args:
        folder:     If folder is set on catalog, add it
        database:   If database is set on catalog, add it
    """
    # Loop through all datasets in the catalog, non-recursive
    for dataset in catalog.get_datasets(recursive=False):
        # If folder input is not None and dataset has no folder already
        if folder and dataset.folder is None:
            dataset_path = join_paths(folder, dataset.name).with_suffix('.' + dataset.data_type)
            setattr(dataset, 'folder', folder)
            setattr(dataset, _PATH_VAR, dataset_path)

        # If database is not None, and dataset has no database already
        if database and dataset.database is None:
            setattr(dataset, 'database', database)



from __future__ import annotations
from pandas import DataFrame


class HTMLDisplay:
    def __init__(self, html_str):
        self.html_str = html_str

    def _repr_html_(self):
        return self.html_str


class DFDisplay(DataFrame):
    def __str__(self):
        return self.to_string(index=False)

    @property
    def _constructor(self) -> DFDisplay:
        """Required property when inheriting a pandas dataframe"""
        return DFDisplay

    def __repr__(self):
        return self.__str__()

    def _repr_html_(self):
        max_chars = 50
        hover_data_df = self.copy().astype(str).applymap(
            lambda x: x.replace('"', "'")
        )
        out_pdf = self.copy().astype(str).applymap(
            lambda x: x if len(x) < max_chars else x[:max_chars - 3] + '...'
        )
        hoverbox_props = (
            'border: 2px solid #000000;'  # Box border size
            'border-radius: 0.4em;'  # Box corners smoothness
            'position:absolute;'  # Where the box should pop up
            'z-index: 1;'  # Where the object should end up (on top)
            'visibility: hidden;'  # The default visibility mode
            'background-color: white;'  # Background color of box
            'color: #000000;'  # Text color
            'font-size: 1.1em;'  # Font size
            'max-width: 80ch;'  # Maximum number of characters per row
            'overflow-wrap: break-word;'  # How to handle text overflow
            'padding: 0.8em;'  # Padding between box and text
            'transform: translate(0px, -24px);'
        )
        return (
            out_pdf
            .style
            .set_table_attributes('class="dataframe"')
            .set_tooltips(hover_data_df, props=hoverbox_props)
            .to_html()
        )



import numpy as np
from collections import Counter
from hela.math.preprocessing import clean_document, clean_corpus
from hela.math.levenshtein import replace
from hela.math.score_info import SimilarityInfo
from typing import Sequence, Tuple


def build_matrix(documents: Sequence[str]) -> Tuple[np.ndarray, Sequence[str]]:
    """Build the tf_idf matrix, returned along with the vocabulary for the matrix.

    Args:
        documents: A list of string documents

    Returns:
        A tuple (X, vocab) as the tf_idf matrix and vocabulary as list
    """
    word_count = [Counter(x.split()) for x in documents]
    vocab = list(set([w for words in word_count for w in words.keys()]))
    arrs = [[wc.get(v, 0) for v in vocab] for wc in word_count]
    X = np.array(arrs)
    idf = np.log(X.shape[1] / np.where(X > 0, 1, 0).sum(axis=0))
    return X * idf, vocab


def normalize(X: np.ndarray) -> np.ndarray:
    """Normalizes each vector in the matrix, i.e. sets vector length = 1.

    Args:
        X: A two-dimensional numpy array

    Returns:
        A normalized numpy matrix
    """
    return X / np.linalg.norm(X, axis=1).reshape(-1, 1)


def cosine_similarity(X: np.ndarray, Y: np.ndarray) -> np.ndarray:
    """Calculates the cosine similarity based on two numpy matrices.

    Args:
        X: A one or two -dimensional numpy array
        Y: A two-dimensional numpy array

    Returns:
        The cosine similarity of each pair between the two matrices.
        Will have length equal to X.shape[0] * Y.shape[0]
    """
    if X.ndim == 1:
        X = X.reshape(1, -1)
    # Set the length of each vector to 1
    X, Y = normalize(X), normalize(Y)
    return (X @ Y.T).flatten()


def find_similar_occurrences(corpus: Sequence[str], min_similarity: float = .75) -> Sequence[SimilarityInfo]:
    """Finds similar occurences of documents in a corpus.

    Args:
        corpus:         A list of string documents
        min_similarity: The minimum required similarity [0 to 1]

    Returns:
        A list of similarity info objects with similarity score >= threshold
    """
    documents, tokenized_documents = clean_corpus(corpus)
    X, _ = build_matrix(tokenized_documents)
    cs = cosine_similarity(X, X).reshape(X.shape[0], -1)
    similarities, match_indices = [], []
    for i, row in enumerate(cs):
        for idx in np.argwhere(row >= min_similarity).flatten():
            # If i == idx that means it is its own match
            # if (idx, i) is in match_indices that means we found the reverse match already
            if i != idx and (idx, i) not in match_indices:
                match_indices.append((i, idx))
                similarities.append(SimilarityInfo(
                    score=round(row[idx], 3),
                    match_idx=idx,
                    match_string=documents[idx],
                    target_idx=i,
                    target_string=documents[i]
                ))
    return sorted(similarities, key=lambda x: -x.score)


def sort(query_str: str, corpus: Sequence[str], fuzzy: bool = True) -> Sequence[SimilarityInfo]:
    """Sort matches based on cosine similarity between query string and corpus.

    Args:
        query_str:  A string of one or more search terms
        corpus:     A list of documents as strings
        fuzzy:      Whether the search terms (query_str) should fuzzily adapt to vocabulary

    Returns:
        A similarity info object with the best possible matching document
    """
    query_dict = Counter(clean_document(query_str).split())
    documents, tokenized_documents = clean_corpus(corpus)
    X, vocab = build_matrix(tokenized_documents)
    if fuzzy:
        query_dict = {
            doc if doc in vocab else replace(doc, vocab): occurrences
            for doc, occurrences in query_dict.items()
        }

    query_arr = np.array([query_dict.get(x, 0) for x in vocab])
    if sum(query_arr) == 0:
        raise ValueError(f'No matching strings found between vocabulary and query string: "{query_str}"')
    sims = cosine_similarity(query_arr, X)
    return [
        SimilarityInfo(
            score=sims[idx],
            match_idx=idx,
            match_string=documents[idx]
        )
        # Argsorting on -sims to get highest similarity first
        for idx in np.argsort(-sims)
    ]



from hela.errors import ValidationError
from hela import Catalog
from hela._catalog_class import is_dataset, is_catalog


def validate_dataset_variable_names(root_catalog: Catalog) -> True:
    """Runs a validation check to make sure field (variable) names of the catalogs
    conform with the name of their datasets.

    Args:
        root_catalog: The root catalog of your project

    Returns:
        True if all datasets were successfully validated

    Raises:
        ValidationError:    Whenever a discrepancy between field (variable) name and dataset
                            name was found
    """
    for field, obj in root_catalog.__dict__.items():
        dataset = is_dataset(obj)
        if dataset:
            if field != dataset.name:
                raise ValidationError(f'Dataset with name "{obj.name}" has field name "{field}"')
        catalog = is_catalog(obj)
        if catalog:
            validate_dataset_variable_names(catalog)
    return True


def validate_no_duplicated_columns(root_catalog: Catalog) -> True:
    """Runs a validation check to make sure there are no duplicated columns among the datasets.

    Args:
        root_catalog: The root catalog of your project

    Returns:
        True if all columns were successfully validated

    Raises:
        ValidationError:    Whenever the same columns if found in multiple datasets
                            without being referenced from store.
    """
    columns_dict = root_catalog.get_columns_datasets()
    for col, datasets in columns_dict.items():
        if len(datasets) > 1 and not col.from_store:
            raise ValidationError(
                f'Datasets {datasets} have overlapping column "{col.name}".'
                f' Consider adding column "{col.name}" to a column store and reference to store column instead.'
            )
    return True



from hela import _constants
from hela._column_classes import _ColumnType
from hela.errors import ValidationError
from collections import Counter


def validate_column_store_field_names(root_store: object) -> True:
    """Runs a validation check on all columns within a column store in order
    to make sure that all store variables have the same name as their
    column name

    Args:
        root_store: A class decorated with @column_store

    Returns:
        True if there were no mismatches

    Raises:
        ValidationError: If any columns are found with mismatching name & field

    Examples:
    >>> @column_store
    >>> class MyStore:
    >>>     b = Col('a', String())
    >>> validate_column_store_field_names(MyStore)
    ValidationError
    >>> @column_store
    >>> class MyStore:
    >>>     a = Col('a', String())
    >>> validate_column_store_field_names(MyStore)
    True
    """
    def _run_check(column_store: object) -> list:
        error_msgs = []
        for field_name, obj in column_store.__dict__.items():
            if isinstance(obj, _ColumnType):
                if field_name != obj.name:
                    error_msgs.append(f'Field "{field_name}" != Column "{obj.name}"')
            elif getattr(obj, _constants._COL_STORE_FLAG, False):
                error_msgs.extend(_run_check(obj))
        return error_msgs

    msgs = _run_check(root_store)
    if len(msgs) == 0:
        return True
    output = '\n'.join(msgs)
    raise ValidationError(f'Found following discrepancies:\n{output}')


def validate_no_duplicate_columns(root_store: object) -> True:
    """Runs a validation check on all columns within the store to make sure there
    are no duplicate columns

    Args:
        root_store: A column store class decorated with @column_store

    Returns:
        True if there are no duplicate columns

    Raises:
        ValidationError: If there exists duplicate columns
    """
    duplicate_columns = []
    for col, occurences in Counter(root_store.columns_recursive_).items():
        if occurences == 1:
            continue
        duplicate_columns.append(f'"{col.name}"')
    if len(duplicate_columns) == 0:
        return True
    out = ', '.join(duplicate_columns)
    raise ValidationError(f'Found duplicates of column(s): {out}')



"""Includes functions to infer Catalog schemas on various data structures."""
from json.decoder import JSONDecodeError
import warnings
import pandas as pd
import json
from typing import List, Union
from pathlib import Path
from collections import defaultdict
from hela.data_types import PrimitiveType, String, Struct, Array
from hela._utils.maps import python_to_data_type_map
from hela import Col
from hela._column_classes import _ColumnType
from hela.errors import InferralError
from datetime import date


def _deep_string_inferral(s: pd.Series) -> PrimitiveType:
    # Attempt to parse column as date
    try:
        return _infer_pandas_column_type(s.apply(lambda x: date.fromisoformat(x)))
    except ValueError:
        pass

    # Attempt to parse column as datetime
    try:
        return _infer_pandas_column_type(pd.to_datetime(s))
    except ValueError:
        pass

    # Attempt to parse column as a json structure (list/dict)
    try:
        return _infer_pandas_column_type(s.apply(lambda x: json.loads(x.replace("'", '"'))))
    except JSONDecodeError:
        pass
    except TypeError:
        pass
    return String()


def _infer_pandas_column_type(s: pd.Series, deep: bool = True) -> PrimitiveType:
    s = s.dropna()

    type_set = set([type(x) for x in s])
    if len(type_set) > 1:
        raise InferralError(f'Could not infer data type for column {s.name}, multiple types found: {type_set}.')
    type_ = list(type_set)[0]

    if deep and type_ == str:
        return _deep_string_inferral(s)

    mapped_type = python_to_data_type_map.get(type_, None)
    if mapped_type is not None:
        return mapped_type

    if type_ == dict:
        combined_dicts = defaultdict(list)
        for d in s:
            for k, v in d.items():
                combined_dicts[k].append(v)
        return Struct({
            k: _infer_pandas_column_type(pd.Series(v, name=k), deep=deep)
            for k, v in combined_dicts.items()
        })

    if type_ == list:
        all_vals = []
        for sub_list in s:
            all_vals.extend(sub_list)
        return Array(_infer_pandas_column_type(pd.Series(sub_list, name=s.name), deep=deep))

    raise InferralError(f'Could not infer data type for column "{s.name}".')


def infer_schema_pandas(df: pd.DataFrame, raise_infer_errors=True, deep=True,
                        sample_size=10**6
                        ) -> List[_ColumnType]:
    """Attempts to infer the types of all columns in a pandas dataframe.

    Can handle nested (dict) columns.

    Args:
        df: A pandas dataframe
        raise_infer_errors: If raise_infer_errors is False, will default to String() and issue a warning
        deep: Will attempt to json load string types
        sample_size: Max number of rows to use for sampling (default 10**6)

    Returns:
        A list of Col objects

    Raises:
        InferralError: When a column or subcolumn could not be inferred

    """
    sample_df = df.sample(min(len(df), sample_size))
    cols = []
    for col, vals in sample_df.items():
        try:
            cols.append(Col(name=col, data_type=_infer_pandas_column_type(vals.copy(), deep=deep)))
        except InferralError as e:
            if raise_infer_errors:
                raise e
            warnings.warn(str(e) + ' Defaulting to String()')
            cols.append(Col(name=col, data_type=String()))
    return cols


def infer_schema_json(path_or_str: Union[Path, str],
                      raise_infer_errors=True, deep=True) -> List[_ColumnType]:
    """Attempts to infer the types of all objects in a json

    Args:
        path_or_str: JSON as string, JSON line (newline separator), Path to JSON, or string path to JSON
        raise_infer_errors: If raise_infer_errors is False, will default to String() and issue a warning
        deep: Will attempt to json load string types

    Returns:
        A list of Col objects

    Raises:
        InferralError: When a column or subcolumn could not be inferred
    """
    if isinstance(path_or_str, str) and path_or_str[-5:] == '.json':
        path_or_str = Path(path_or_str).read_text()
    elif isinstance(path_or_str, Path):
        path_or_str = path_or_str.read_text()

    try:
        dict_list = json.loads(path_or_str)
    except JSONDecodeError:
        dict_list = [json.loads(x) for x in path_or_str.split('\n')]

    try:
        df = pd.DataFrame(dict_list)
    except ValueError:
        df = pd.DataFrame([dict_list])

    return infer_schema_pandas(df, raise_infer_errors=raise_infer_errors, deep=deep)



import numpy as np
import pandas as pd
from hela.data_types import String, Double, Long, DateTime, Date, Bool
from datetime import date, datetime


python_to_data_type_map = {
    str: String(),
    int: Long(),
    float: Double(),
    bool: Bool(),
    datetime: DateTime(),
    date: Date(),
    np.datetime64: DateTime(),
    pd.Timestamp: DateTime()
}



"""Module consisting of all pre-built data types."""
import json
from abc import abstractmethod
from hela._base_data_type import BaseDataType
from hela._utils.bigquery_utils import BigqueryType, BigqueryMode
from typing import Dict

try:
    from aws_cdk.aws_glue import Schema as GlueSchema, Column as GlueColumn
except ModuleNotFoundError:
    pass
except json.decoder.JSONDecodeError:
    pass

try:
    from pyspark.sql import types as spark_types
except ModuleNotFoundError:
    pass

try:
    from google.cloud import bigquery
except ModuleNotFoundError:
    pass


class PrimitiveType(BaseDataType):
    """Helper class that sets up abstract methods for all other data_type classes.

    Attributes:
        name:   The name of the data type.
        description:    A description of the data type.
        json_kwargs:    Optional json key-value arguments to use for generating json schemas.
    """

    def __init__(self, name: str, description: str, json_kwargs: dict = None) -> None:
        super().__init__(name=name, description=description)
        self.json_kwargs = {} if json_kwargs is None else json_kwargs

    @abstractmethod
    def _spark_type(self):
        raise NotImplementedError

    @abstractmethod
    def _glue_type(self):
        raise NotImplementedError

    @abstractmethod
    def _json_type(self):
        raise NotImplementedError

    @abstractmethod
    def _bigquery_type(self) -> BigqueryType:
        raise NotImplementedError


class String(PrimitiveType):
    """Regular string data type.

    Attributes:
        json_kwargs:    Optional json key-value arguments to use for generating json schemas.
    """

    def __init__(self, json_kwargs: dict = None) -> None:
        super().__init__('String', 'String object', json_kwargs=json_kwargs)

    def _spark_type(self):
        return spark_types.StringType()

    def _glue_type(self):
        return GlueSchema.STRING

    def _json_type(self):
        return {'type': 'string', **self.json_kwargs}

    def _bigquery_type(self) -> BigqueryType:
        return BigqueryType('STRING')


class Int(PrimitiveType):
    """Integer data type, represented by 32 bits.

    Attributes:
        json_kwargs:    Optional json key-value arguments to use for generating json schemas.
    """

    def __init__(self, json_kwargs: dict = None) -> None:
        super().__init__('Int', 'Integer', json_kwargs=json_kwargs)

    def _spark_type(self):
        return spark_types.IntegerType()

    def _glue_type(self):
        return GlueSchema.INTEGER

    def _json_type(self):
        return {'type': 'integer', **self.json_kwargs}

    def _bigquery_type(self) -> BigqueryType:
        return BigqueryType('INT64')


class Long(PrimitiveType):
    """Long numerical data type, represented by 64 bits.

    Attributes:
        json_kwargs:    Optional json key-value arguments to use for generating json schemas.
    """

    def __init__(self, json_kwargs: dict = None) -> None:
        super().__init__('Long', 'Long', json_kwargs=json_kwargs)

    def _spark_type(self):
        return spark_types.LongType()

    def _glue_type(self):
        return GlueSchema.BIG_INT

    def _json_type(self):
        return {'type': 'integer', **self.json_kwargs}

    def _bigquery_type(self) -> BigqueryType:
        return BigqueryType('INT64')


class Float(PrimitiveType):
    """Float decimal data type, usually represented by 32 bits.

    Attributes:
        json_kwargs:    Optional json key-value arguments to use for generating json schemas.
    """

    def __init__(self, json_kwargs: dict = None) -> None:
        super().__init__('Float', 'Float', json_kwargs=json_kwargs)

    def _spark_type(self):
        return spark_types.FloatType()

    def _glue_type(self):
        return GlueSchema.FLOAT

    def _json_type(self):
        return {'type': 'number', **self.json_kwargs}

    def _bigquery_type(self) -> BigqueryType:
        return BigqueryType('FLOAT64')


class Double(PrimitiveType):
    """Double decimal data type, usually represented by 64 bits.

    Attributes:
        json_kwargs:    Optional json key-value arguments to use for generating json schemas.
    """

    def __init__(self, json_kwargs: dict = None) -> None:
        super().__init__('Double', 'Double', json_kwargs=json_kwargs)

    def _spark_type(self):
        return spark_types.DoubleType()

    def _glue_type(self):
        return GlueSchema.DOUBLE

    def _json_type(self):
        return {'type': 'number', **self.json_kwargs}

    def _bigquery_type(self) -> BigqueryType:
        return BigqueryType('FLOAT64')


class Date(PrimitiveType):
    """Date data type e.g. YYYY-MM-dd

    Attributes:
        json_kwargs:    Optional json key-value arguments to use for generating json schemas.
    """

    def __init__(self, json_kwargs: dict = None) -> None:
        super().__init__('Date', 'Date', json_kwargs=json_kwargs)

    def _spark_type(self):
        return spark_types.DateType()

    def _glue_type(self):
        return GlueSchema.DATE

    def _json_type(self):
        return {'type': 'string', 'format': 'date', **self.json_kwargs}

    def _bigquery_type(self) -> BigqueryType:
        return BigqueryType('DATE')


class DateTime(PrimitiveType):
    """Datetime data type.

    Attributes:
        json_kwargs:    Optional json key-value arguments to use for generating json schemas.
    """

    def __init__(self, json_kwargs: dict = None) -> None:
        super().__init__('DateTime', 'DateTime', json_kwargs=json_kwargs)

    def _spark_type(self):
        return spark_types.TimestampType()

    def _glue_type(self):
        return GlueSchema.TIMESTAMP

    def _json_type(self):
        return {'type': 'string', 'format': 'date-time', **self.json_kwargs}

    def _bigquery_type(self) -> BigqueryType:
        return BigqueryType('DATETIME')


class Bool(PrimitiveType):
    """Bool data type.

    Attributes:
        json_kwargs:    Optional json key-value arguments to use for generating json schemas.
    """

    def __init__(self, json_kwargs: dict = None) -> None:
        super().__init__('Bool', 'Boolean', json_kwargs=json_kwargs)

    def _spark_type(self):
        return spark_types.BooleanType()

    def _glue_type(self):
        return GlueSchema.BOOLEAN

    def _json_type(self):
        return {'type': 'boolean', **self.json_kwargs}

    def _bigquery_type(self) -> BigqueryType:
        return BigqueryType('BOOL')


class Array(PrimitiveType):
    """Array data type, expects another primitive data type upon initialization.

    Attributes:
        item_type:  Another data type held within the array. All sub items must be of the same type.
        json_kwargs:    Optional json key-value arguments to use for generating json schemas.

    Examples:
    >>> from hela.data_types import Array, String
    >>> Array(String())
    """

    def __init__(self, item_type: PrimitiveType, json_kwargs: dict = None) -> None:
        super().__init__(f'Array({item_type.name})', 'No Description.', json_kwargs=json_kwargs)
        self.item_type = item_type

    def _spark_type(self):
        return spark_types.ArrayType(self.item_type._spark_type())

    def _glue_type(self):
        sub_type = self.item_type._glue_type()  # Returns e.g. GlueSchema.INTEGER
        return GlueSchema.array(
            input_string=sub_type.input_string,
            is_primitive=sub_type.is_primitive
        )

    def _json_type(self):
        return {'type': 'array', 'items': self.item_type._json_type(), **self.json_kwargs}

    def _bigquery_type(self) -> BigqueryType:
        array_type = self.item_type._bigquery_type()
        array_type.mode = BigqueryMode.REPEATED
        return array_type


class Struct(PrimitiveType):
    """Struct/dict/record data type, expects a dict of {str: PrimitiveType}.

    For nested columns see `hela.NestedCol`.

    Attributes:
        nested_types:   A dictionary of str: Primitive type.
        json_kwargs:    Optional json key-value arguments to use for generating json schemas.

    Examples:
    >>> from hela.data_types import Struct, String, Int
    >>> Struct({'my_string': String(), 'my_int': Int()})
    """

    def __init__(self, nested_types: Dict[str, PrimitiveType], json_kwargs: dict = None) -> None:
        desc_repr = {name: type_.name for name, type_ in nested_types.items()}
        super().__init__(f'Struct({desc_repr})', description='No Description.', json_kwargs=json_kwargs)
        self.nested_types = nested_types

    def _spark_type(self):
        return spark_types.StructType([
            spark_types.StructField(name=name, dataType=type_._spark_type())
            for name, type_ in self.nested_types.items()
        ])

    def _glue_type(self):
        return GlueSchema.struct([
            GlueColumn(name=name, type=type_._glue_type())
            for name, type_ in self.nested_types.items()
        ])

    def _json_type(self):
        nested_json_types = {name: type_._json_type() for name, type_ in self.nested_types.items()}
        return {'type': 'object', 'properties': nested_json_types, 'additionalProperties': False, **self.json_kwargs}

    def _bigquery_type(self):
        fields = (
            bigquery.SchemaField(name, **type_._bigquery_type().__dict__)
            for name, type_ in self.nested_types.items()
        )
        return BigqueryType(
            'RECORD',
            mode=BigqueryMode.NULLABLE,
            fields=fields
        )



import pandas as pd
from hela import BaseDataset
from hela._column_classes import _ColumnType
from typing import Optional, Sequence, Set, Union, Dict, Any
from datetime import date
from pathlib import Path


class PandasParquetDataset(BaseDataset):
    def __init__(
        self,
        name: str,
        folder: Optional[Union[str, Path]] = None,
        description: Optional[str] = None,
        rich_description_path: Optional[str] = None,
        partition_cols: Optional[Sequence[str]] = None,
        columns: Optional[Sequence[_ColumnType]] = None
    ) -> None:
        super().__init__(
            name=name,
            data_type='parquet',
            folder=folder,
            description=description,
            rich_description_path=rich_description_path,
            columns=columns,
            partition_cols=partition_cols
        )

    def write(self, df: pd.DataFrame) -> int:
        df.to_parquet(self.path, partition_cols=self.partition_cols)
        return len(df)

    def load(self) -> pd.DataFrame:
        if self.path is None:
            raise FileNotFoundError('No path has been specified for this dataset.')
        return pd.read_parquet(self.path)

    def get_dates(self) -> Optional[Set[date]]:
        if self.path is None:
            return None
        if not self.path.exists():
            return None
        return set([
            date.fromisoformat(p.stem.split('=')[-1])
            for p in self.path.iterdir()
        ])

    def get_samples(self) -> Optional[Dict[str, Any]]:
        try:
            return self.load().apply(lambda x: x[x.first_valid_index()], axis=0).to_dict()
        except FileNotFoundError:
            return None



import gzip
import pkg_resources
from pathlib import Path
from typing import Union, Sequence
from hela import Catalog
from hela.web_page._json_generator import JsonGenerator


def generate_webpage(
    catalogs: Union[Catalog, Sequence[Catalog]],
    output_folder: str,
    overwrite_existing: bool = False,
    include_samples: bool = False,
    web_app_title: str = None
) -> None:
    """Generates an index.html file that can be used as a data catalog website.

    Include a python script implementing this function in your CI/CD pipeline, outputting an index.html file
    that you can then use to share your data catalog (e.g. on github pages).
    For an example see (TODO: insert example repo link here).

    Args:
        catalogs:   One or multiple objects inheriting the Catalog class.
                    If you have a tree of catalogs, only the root catalog is required.
        output_folder:  The folder where index.html file should end up.
        overwrite_existing: Flag whether and potential index.html file should be overwritten if existing.
        include_samples:    Flag whether to attempt to fetch sample datapoints from the columns in each
                            dataset. Requires `hela.BaseDataset.get_samples` function implemented.
        web_app_title:  Optional title of the web app.

    Raises:
        FileExistsError: If the index.html file already exists under `output_folder` and overwrite_existing=False.

    Examples:
    >>> from my_catalog import MyCatalog
    >>> from hela import generate_webpage
    >>> generate_webpage(MyCatalog, '.', overwrite_existing=True)
    """
    if not isinstance(catalogs, Sequence):
        catalogs = [catalogs]

    jg = JsonGenerator()
    json_str = jg.generate_docs_jsons(catalogs, include_samples=include_samples)
    folder_path = Path(output_folder)
    file_path = folder_path / 'index.html'
    if not folder_path.exists():
        folder_path.mkdir(parents=True)

    if file_path.exists():
        if not overwrite_existing:
            raise FileExistsError(f'File {file_path} already exists, delete or set overwrite_existing=True')
        file_path.unlink()

    # Replace placeholder script with actual json data
    replacement_str = f'<script>window.treeListData = {json_str}</script>'
    match_str = '<script id="tree-list-data"></script>'
    original_index_file = gzip.decompress(pkg_resources.resource_string(__name__, 'index.html.gz')).decode()
    output_file = original_index_file.replace(match_str, replacement_str)

    # Replace web app title with custom title
    if web_app_title:
        output_file = output_file.replace('<title>Dashboard</title>', f'<title>{web_app_title}</title>')
    file_path.write_text(output_file)



"""
.. include:: ../gh_pages/hela.md
"""

from hela._catalog_class import Catalog
from hela._column_classes import Col, NestedCol
from hela._base_dataset import BaseDataset
from hela._column_store_class import column_store
from hela.web_page.generate import generate_webpage


__all__ = [
    'Catalog',
    'BaseDataset',
    'column_store',
    'Col',
    'NestedCol',
    'generate_webpage'
]



from hela import Catalog
from hela.math.tf_idf import find_similar_occurrences
from hela.math.preprocessing import clean_document
from hela.errors import ValidationError
from collections import defaultdict


def validate_description_similarity(root_catalog: Catalog, min_similarity: float = .75) -> True:
    """Runs a validation check to make sure there descriptions too similar within the catalog.

    This check does not consider duplicated descriptions, for that use validate_no_duplicated_descriptions().

    Args:
        root_catalog: The root catalog of your project
        max_similarity: The minimum required similarity score before raising a validation error (0 to 1)

    Returns:
        True if no description was too similar to another.

    Raises:
        ValidationError:    Raised when descriptions were found that too closely resembles each other.
    """

    desc_dict = {sd.description: sd for sd in root_catalog._all_descriptions()}

    similar_descriptions = find_similar_occurrences(list(desc_dict.keys()), min_similarity=min_similarity)
    msgs = []
    for sd in similar_descriptions:
        obj1 = desc_dict[sd.target_string]
        obj2 = desc_dict[sd.match_string]
        msgs.append(
            f'{obj1.type} <{obj1.name}> too closely resembles {obj2.type} <{obj2.name}>'
            f' ({round(sd.score, 2)}), descriptions:\n'
            f'<{obj1.name}> {obj1.description}\n<{obj2.name}> {obj2.description}'
        )
    if msgs:
        msgs = ['Found descriptions too close in similarity:'] + msgs
        raise ValidationError('\n'.join(msgs))
    return True


def validate_no_description_duplication(root_catalog: Catalog) -> True:
    """Runs a validation check to make sure no descriptions are duplicated within the catalog.

    Args:
        root_catalog: The root catalog of your project

    Returns:
        True if no descriptions were duplicated

    Raises:
        ValidationError: If any duplicated descriptions were found.
    """

    desc_dict = defaultdict(list)
    for sd in root_catalog._all_descriptions():
        if sd is not None:
            desc_dict[clean_document(sd.description)].append(sd)

    msgs = []
    for description, objects in desc_dict.items():
        if len(objects) > 1:
            obj_list = [f'{obj.name} ({obj.type})' for obj in objects]
            msgs.append(
                f'The following objects share the (tokenized) description <{description}>: {", ".join(obj_list)}'
            )
    if msgs:
        msgs = ['Duplicated descriptions found:'] + msgs
        raise ValidationError('\n'.join(msgs))
    return True



from datetime import date, timedelta
from typing import Set


def get_date_range_from_set(dates: Set[date]) -> Set[date]:
    """
        Returns an inclusive set range with all available dates between max and min of input dates.

        Args:
            dates: A set of dates e.g. {date(2020, 1, 1), date(202, 1, 3)}

        Returns:
            An inclusive set of dates [min_date, max_date]
    """
    min_date, max_date = min(dates), max(dates)
    return {min_date + timedelta(days=x) for x in range((max_date - min_date).days + 1)}


def get_missing_dates(dates: Set[date]) -> Set[date]:
    """
        Returns dates missing from dataset as a set of dates.
        Requires .get_dates() to be implemented.

        Returns:
            Set of missing dates
    """
    return get_date_range_from_set(dates) - dates



"""Module with custom errors."""


class InferralError(Exception):
    """Error sent when an inferral failed."""
    pass


class DatasetError(Exception):
    pass


class DuplicationError(Exception):
    pass


class ValidationError(Exception):
    pass



from hela import BaseDataset
from hela._column_classes import _ColumnType
from datetime import date
from typing import Optional, Union, Sequence, Set
from pathlib import Path
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.functions import first


def spark_session() -> SparkSession:
    return (
        SparkSession.builder
        .appName('testing')
        .master('local[2]')
        .getOrCreate()
    )


class SparkParquetDataset(BaseDataset):
    def __init__(
        self,
        name: str,
        folder: Optional[Union[str, Path]] = None,
        description: Optional[str] = None,
        rich_description_path: Optional[str] = None,
        partition_cols: Optional[Sequence[str]] = None,
        columns: Optional[Sequence[_ColumnType]] = None
    ) -> None:
        super().__init__(
            name=name,
            data_type='parquet',
            folder=folder,
            description=description,
            rich_description_path=rich_description_path,
            columns=columns,
            partition_cols=partition_cols
        )

    def spark(self) -> SparkSession:
        return spark_session()

    def write(self, df: DataFrame) -> int:
        df.write.mode('overwrite').parquet(str(self.path), partitionBy=self.partition_cols)
        return df.count()

    def load(self):
        return self.spark().read.parquet(str(self.path))

    def get_dates(self) -> Set[date]:
        return set([
            date.fromisoformat(p.stem.split('=')[-1])
            for p in self.path.iterdir()
            if len(p.stem.split('=')) > 1
        ])

    def get_samples(self) -> dict:
        df = self.load().limit(1000)
        return df.select([first(x, ignorenulls=True).alias(x) for x in df.columns]).first().asDict()



from dataclasses import dataclass, field


@dataclass
class BigqueryMode:
    NULLABLE = 'NULLABLE'
    REPEATED = 'REPEATED'


@dataclass
class BigqueryType:
    """Bigquery type, used to standardise the way bigquery Data Types are propagated down."""
    field_type: str
    mode: BigqueryMode = BigqueryMode.NULLABLE
    fields: tuple = field(default_factory=tuple)



from __future__ import annotations
from abc import abstractmethod, ABC
from copy import copy
from dataclasses import dataclass
from hela._base_data_type import BaseDataType
from hela.data_types import Struct, PrimitiveType
from hela._utils.bigquery_utils import BigqueryMode
from typing import Optional, Sequence

from hela._utils.short_description import ShortDescription

try:
    from aws_cdk.aws_glue import Column as GlueColumn
except ModuleNotFoundError:
    pass

try:
    from pyspark.sql.types import StructField
except ModuleNotFoundError:
    pass

try:
    from google.cloud import bigquery
except ModuleNotFoundError:
    pass


@dataclass
class _ColInfo:
    name: str
    data_type: str
    description: str
    from_store: bool


class _ColumnType(BaseDataType, ABC):
    _type: str = 'Column'
    data_type: BaseDataType
    label: Optional[str] = None
    from_store: bool = False

    def __init__(self, name: str, data_type: BaseDataType, description: str) -> None:
        super().__init__(name=name, description=description)
        self.data_type = data_type

    def __eq__(self, o: _ColumnType) -> bool:
        return self.__hash__() == o.__hash__()

    def __hash__(self) -> int:
        return hash(str(self.name) + str(self.data_type))

    def __str__(self) -> str:
        return f'Col:{self.name}'

    def __repr__(self) -> str:
        return self.__str__()

    @abstractmethod
    def _describe(self) -> Sequence[_ColInfo]:
        raise NotImplementedError

    @abstractmethod
    def _spark_type(self):
        raise NotImplementedError

    @abstractmethod
    def _glue_type(self):
        raise NotImplementedError

    @abstractmethod
    def _json_type(self):
        raise NotImplementedError

    @abstractmethod
    def _bigquery_type(self):
        raise NotImplementedError

    def _desc_(self) -> ShortDescription:
        return ShortDescription(name=self.name, type=self._type, description=self.description)


class Col(_ColumnType):
    """A basic column object, for nested columns see `hela.NestedCol`.

    This class is used to define columns within a `hela.BaseDataset` or `hela.column_store`.
    Each defined column will be searchable, testable and referenceable.

    If you want to give further functionality to your columns, please this class.

    Attributes:
        name:   The name of the column.
        data_type:  The data type of the column, should be one of types found in `hela.data_types`
        description: A description of this column as a string, better descriptions yield a more secure catalog.

    Examples:
    >>> from hela import Col
    >>> from hela.data_types import String
    >>> my_col = Col('my_col', String(), 'This is an example column')
    """

    def __init__(
        self,
        name: str,
        data_type: PrimitiveType,
        description: str = None,
    ) -> None:
        super().__init__(name=name, data_type=data_type, description=description)

    def _describe(self) -> Sequence[_ColInfo]:
        return [_ColInfo(
            name=self.name,
            data_type=str(self.data_type),
            description=self.description,
            from_store=self.from_store
        )]

    def _spark_type(self):
        return StructField(name=self.name, dataType=self.data_type._spark_type())

    def _glue_type(self):
        return GlueColumn(name=self.name, type=self.data_type._glue_type(), comment=self.description)

    def _json_type(self):
        return {self.name: self.data_type._json_type()}

    def _bigquery_type(self):
        return bigquery.SchemaField(
            name=self.name,
            description=self.description,
            **self.data_type._bigquery_type().__dict__
        )

    def __str__(self) -> str:
        return f'Col(name="{self.name}", data_type={self.data_type}, description={self.description})'

    def __repr__(self) -> str:
        return self.__str__()


class NestedCol(_ColumnType):
    """A nested style column object, should be instantiated with sub columns.

    Most data stores support nested style columns, these can be built using this column class.
    These columns will be referenced with dot-notation when shown in the **catalog**.
    For dict/struct style columns see `hela.data_types.Struct`.

    Attributes:
        name:   The name of the column.
        columns:    A sequence of columns nested within this column. Can be Col or NestedCol objects.

    Examples:
    >>> from hela import NestedCol, Col
    >>> from hela.data_types import String, Int
    >>> my_col = NestedCol('my_nested_col', [
    ...     Col('nested_string', String(), 'Nested string column'),
    ...     Col('nested_int', Int(), 'Nested int column')
    ... ])
    """

    def __init__(self, name: str, columns: Sequence[_ColumnType]) -> None:
        self.columns = columns
        data_type = Struct({c.name: c.data_type for c in columns})
        super().__init__(name=name, data_type=data_type, description='A subset of columns.')

    def _describe(self) -> Sequence[_ColInfo]:
        col_info_list = []
        for c in self.columns:
            for c_desc in c._describe():
                desc = copy(c_desc)
                desc.name = f'{self.name}.{desc.name}'
                col_info_list.append(desc)
        return col_info_list

    def _spark_type(self):
        return StructField(name=self.name, dataType=self.data_type._spark_type())

    def _glue_type(self):
        return GlueColumn(name=self.name, type=self.data_type._glue_type())

    def _json_type(self):
        return {self.name: self.data_type._json_type()}

    def _bigquery_type(self):
        return bigquery.SchemaField(
            name=self.name,
            field_type='RECORD',
            mode=BigqueryMode.NULLABLE,
            fields=[c._bigquery_type() for c in self.columns]
        )

    def __str__(self) -> str:
        subs = [str(c) for c in self.columns]
        return f'NestedCol(name="{self.name}", subcols={subs})'

    def _desc_(self) -> Sequence[ShortDescription]:
        desc_list = []
        for c_desc in self.columns:
            desc = copy(c_desc._desc_())
            desc.name = f'{self.name}.{c_desc.name}'
            desc_list.append(desc)
        return desc_list

    def __repr__(self) -> str:
        return self.__str__()



from abc import ABC, abstractmethod


class BaseDataType(ABC):
    def __init__(self, name, description) -> None:
        self.name = name
        self.description = description

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.__str__()

    @abstractmethod
    def _spark_type(self):
        raise NotImplementedError

    @abstractmethod
    def _glue_type(self):
        raise NotImplementedError

    @abstractmethod
    def _json_type(self):
        raise NotImplementedError

    @abstractmethod
    def _bigquery_type(self):
        raise NotImplementedError

    def _desc_(self) -> str:
        return {'name': self.name, 'type': 'BaseDataType', 'desc': self.description}



import os
import pandas as pd
from typing import Optional, Sequence
from hela import BaseDataset
from hela._column_classes import _ColumnType
from hela.schema_generators import bigquery_schema
from hela._utils.dict_utils import flatten_dict
try:
    from google.cloud import bigquery
except ModuleNotFoundError:
    pass


class BigqueryDataset(BaseDataset):
    def __init__(
        self,
        name: str,
        database: Optional[str] = None,
        description: Optional[str] = None,
        rich_description_path: Optional[str] = None,
        partition_cols: Optional[Sequence[str]] = None,
        columns: Optional[Sequence[_ColumnType]] = None
    ) -> None:
        super().__init__(
            name,
            data_type='bigquery',
            folder=None,
            database=database,
            description=description,
            rich_description_path=rich_description_path,
            partition_cols=partition_cols,
            columns=columns
        )
        self.client = bigquery.Client()

    @property
    def project_id(self) -> str:
        """Returns the bigquery project id, assumes project_id exists under env variable "CLOUDSDK_CORE_PROJECT"""
        return os.environ['CLOUDSDK_CORE_PROJECT']

    @property
    def table_id(self) -> str:
        """Returns the full table id in style of: "project_id.database_id.table_id"
        Assumes project_id exists under env variable "CLOUDSDK_CORE_PROJECT"
        """
        return f'{self.project_id}.{self.database}.{self.name}'

    def write_json(self, json_rows: Sequence[dict]) -> None:
        self.client.insert_rows_json(self.table_id, json_rows)

    def query_table(self, query_string: str) -> pd.DataFrame:
        """Query this dataset in bigquery. In the FROM part simply write TABLE.
        >>> SELECT * FROM TABLE
        instead of
        >>> SELECT * FROM <some_table_id>

        Args:
            query_string: A query for this dataset/table
        """
        query_string = query_string.replace('FROM TABLE', f'FROM {self.table_id}')

        return self.client.query(query_string).to_dataframe()

    def create_table(self) -> None:
        """Create a bq table given this datasets column schema."""
        table = bigquery.Table(
            self.table_id,
            schema=bigquery_schema(self.columns)
        )
        self.client.create_table(table)

    def create_database(self) -> None:
        """Create a bq database given this datasets database (or its catalog's database)"""
        if self.database is None:
            raise ValueError('Database is None')
        self.client.create_dataset(f'{self.project_id}.{self.database}')

    def delete_table(self) -> None:
        """Delete this table from bigquery"""
        self.client.delete_table(self.table_id)

    def get_samples(self) -> dict:
        df = self.query_table('SELECT * FROM TABLE LIMIT 1')
        sample_dict = df.to_dict(orient='records')[0]
        # Return both original and flattened keys to make sure we have both
        # my_column {a: 1, b: 2}
        # and
        # my_column.a
        # my_column.b
        return {
            **sample_dict,
            **flatten_dict(sample_dict)
        }



"""Module used to translate from catalog schema to other schema types."""
from hela._column_classes import _ColumnType
from hela import BaseDataset
from typing import Sequence, Union


def _get_columns(obj: Union[Sequence[_ColumnType], BaseDataset]) -> Sequence[_ColumnType]:
    if isinstance(obj, BaseDataset):
        obj = obj.columns
    if len(obj) == 0:
        raise ValueError('The column list is empty.')
    return obj


def spark_schema(obj: Union[Sequence[_ColumnType], BaseDataset]):
    col_list = _get_columns(obj)
    from pyspark.sql.types import StructType
    return StructType([c._spark_type() for c in col_list])


def aws_glue_schema(obj: Union[Sequence[_ColumnType], BaseDataset]):
    col_list = _get_columns(obj)
    return [c._glue_type() for c in col_list]


def bigquery_schema(obj: Union[Sequence[_ColumnType], BaseDataset]):
    col_list = _get_columns(obj)
    return [c._bigquery_type() for c in col_list]


def json_schema(obj: Union[Sequence[_ColumnType], BaseDataset]):
    col_list = _get_columns(obj)
    properties = {}
    for c in col_list:
        properties.update(c._json_type())
    return {
        '$schema': 'http://json-schema.org/draft-07/schema',
        'type': 'object',
        'properties': properties
    }



from hela._column_classes import _ColumnType
from hela._constants import _COL_STORE_FLAG


def _make_column_store(cls: object, label: str = None):
    columns_, columns_recursive_ = [], []
    for field, obj in cls.__dict__.items():
        if field.startswith('__') and field.endswith('__'):
            continue

        if getattr(obj, _COL_STORE_FLAG, False):
            columns_recursive_.extend(obj.columns_recursive_)
            continue

        if issubclass(obj.__class__, _ColumnType):
            if obj.name is None:
                obj.name = field
            obj.from_store = True
            obj.label = label
            columns_.append(obj)

    # Set properties to the column store
    setattr(cls, 'columns_', columns_)
    setattr(cls, 'columns_recursive_', columns_ + columns_recursive_)

    # Add functions to column store
    setattr(cls, _COL_STORE_FLAG, True)

    # Sort all fields in the column store alphabetically, for easier auto-complete
    all_fields = [(field_name, field_obj) for field_name, field_obj in cls.__dict__.items()]
    for f_tuple in sorted(all_fields, key=lambda x: x[0], reverse=True):
        # Cannot set __dict__, skip it
        if f_tuple[0] == '__dict__':
            continue
        setattr(cls, f_tuple[0], f_tuple[1])
    return cls


def column_store(cls=None, label: str = None) -> object:
    """Decorator to used to flag a class as a column store.

    A column store is a referencable class used when multiple datasets use the same column.
    In order to ensure that this column is purposefully duplicated among datasets we
    check that any duplicated column must originate from the same column store.

    Args:
        label: This string label will be passed down to all column objects within the store.

    Returns:
        The decorated class.


    Examples:
        >>> from hela import column_store, Col
        >>> from hela.data_types import String
        >>> @column_store(label='cool_columns')
        >>> class MyStore:
        ...     my_column = Col('my_column', String(), 'Example column')
        >>> MyStore.my_column
    """

    def wrap(cls):
        return _make_column_store(cls, label=label)

    # This is triggered when called as @column_store() (with parentheses)
    if cls is None:
        return wrap

    # This is triggered when called as @column_store (no parentheses)
    # Only allow triggered with parenthesis, this will keep IDE hints.
    raise ValueError('A column store can only be decorated with called method: `@column_store()`')



from dataclasses import dataclass
from typing import Optional


@dataclass
class SimilarityInfo:
    score: float
    match_idx: int
    match_string: str
    target_string: Optional[str] = None
    target_idx: Optional[str] = None



"""
Module for plot functions.
"""
from .date_plots import plot_date_availability_calendar
from .html_displays import HTMLDisplay, DFDisplay

__all__ = [
    'plot_date_availability_calendar',
    'HTMLDisplay',
    'DFDisplay'
]



def flatten_dict(d: dict, parent: str = None) -> dict:
    r = {}
    for k, v in d.items():
        if isinstance(v, dict):
            kp = k if parent is None else parent + '.' + k
            r.update(flatten_dict(v, kp))
        else:
            if parent is None:
                r[k] = v
            else:
                r[f'{parent}.{k}'] = v
    return r



import re
import pkg_resources
from typing import Sequence, Tuple


# Load stopwords when module is loaded
stopwords = (
    pkg_resources
    .resource_string(__name__, "stopwords.txt")
    .decode()
    .strip()
    .split()
)


def clean_document(document: str) -> str:
    # Replace all space-like characters with spaces
    document = re.sub("[!,?._@+/-]", " ", document)
    # Remove all non-characters
    document = re.sub("[^a-z\\s]", "", document.lower())
    return ' '.join([x for x in document.split() if x not in stopwords])


def clean_corpus(corpus: Sequence[str]) -> Tuple[Sequence[str], Sequence[str]]:
    """Cleans a corpus of documents, returning original without invalid documents and cleaned corpus.

    Args:
        corpus: A sequence of string documents

    Returns:
        Tuple of:
            Original corpus, with any invalid documents removed
            Cleaned tokenized corpus
    """

    cleaned_corpus, tokenized_corpus = [], []
    for doc in corpus:
        if doc is None:
            continue
        clean_doc = clean_document(doc)
        if clean_doc:
            cleaned_corpus.append(doc)
            tokenized_corpus.append(clean_doc)
    return cleaned_corpus, tokenized_corpus



import json
import warnings
from pathlib import Path
from hela import Catalog, BaseDataset
from typing import Dict, Any, Sequence
from collections import defaultdict


class JsonGenerator:
    _columns_dict: Dict[str, Sequence[BaseDataset]] = None

    def _build_dataset_columns_json(
        self, dataset: BaseDataset, include_samples: bool
    ) -> Sequence[Dict[str, Any]]:
        dataset_columns = dataset.columns
        columns_list = []
        # Attempt to fetch samples if boolean flag is set to true
        try:
            samples = dataset.get_samples() if include_samples else None
        except NotImplementedError:
            warnings.warn(
                f'get_samples() function not implemented for dataset {dataset.name} when include_samples=True'
            )
            samples = None
        samples = {} if samples is None else samples

        # If there are no columns in the dataset
        if dataset_columns is None:
            return []

        # Loop through the columns and build info json
        for col in dataset.columns:
            for col_info in col._describe():
                other_datasets = [
                    {'name': d.name, 'id': d._id}
                    for d in self._columns_dict[col_info.name]
                    if d._id != dataset._id
                ]
                sample = samples.get(col_info.name, None)
                sample = None if sample is None else str(sample)
                columns_list.append({
                    **col_info.__dict__,
                    'sample_data': sample,
                    'other_dataset': other_datasets
                })
        return columns_list

    def _load_rich_description(self, path_str):
        if path_str is None:
            return None
        path = Path(path_str)
        if path.exists():
            return Path(path_str).read_text()
        return f'Rich description could not be found under path {path}'

    def _build_catalog_descriptions(self,
                                    catalog: Catalog,
                                    parent: str = None,
                                    include_samples: bool = False
                                    ) -> Dict[str, Any]:
        child_list = []
        catalog_id_str = catalog.__name__ if parent is None else parent + '--' + catalog.__name__
        for cat in catalog.get_catalogs(recursive=False):
            child_list.append(self._build_catalog_descriptions(cat, catalog_id_str))

        for dataset in catalog.get_datasets(recursive=False):
            column_data = self._build_dataset_columns_json(dataset, include_samples=include_samples)
            child_list.append(
                {
                    **dataset._desc_().__dict__,
                    'rich_description': self._load_rich_description(dataset.rich_description_path),
                    'id': dataset._id,
                    'columns': column_data,
                }
            )

        return {
            **catalog._desc_().__dict__,
            'id': catalog_id_str,
            'rich_description': self._load_rich_description(catalog._rich_description_path),
            'children': child_list
        }

    def generate_docs_jsons(self, catalogs: Sequence[Catalog], include_samples: bool):
        # Instantiate columns dict with columns from all datasets in all catalogs
        columns_dict = defaultdict(set)
        for catalog in catalogs:
            for col, datasets in catalog.get_columns_datasets().items():
                columns_dict[col.name].update(datasets)
        self._columns_dict = columns_dict

        json_str = json.dumps([
            self._build_catalog_descriptions(c, include_samples=include_samples)
            for c in catalogs
        ])
        return json_str



from typing import Sequence
import pandas as pd
import datetime
from hela._utils.date_utils import get_date_range_from_set
from hela.plots.html_displays import HTMLDisplay


def plot_date_availability_calendar(dates: Sequence[datetime.date]) -> HTMLDisplay:
    colors = {
        'exists': '#77DD77',
        'missing': '#FF6961',
        'NA': 'white'
    }

    def color_df(val):
        try:
            day, exists = val
        except TypeError:
            day, exists = '', None
        color = colors['NA'] if exists is None else colors['exists'] if exists else colors['missing']
        return f'c:{color}{day}'

    possible_dates = list(get_date_range_from_set(dates))
    date_check_list = list(zip(possible_dates, [True if d in dates else False for d in possible_dates]))
    df = pd.DataFrame([
        {'year-month': (d.year, d.month), 'date': d.day, 'day_exists': (d.day, exists)}
        for d, exists in date_check_list
    ])
    df = df.pivot(index='year-month', columns='date')
    df = df.applymap(color_df)
    html_str = df.to_html(escape=False, header=False)
    replacements = [(f'<td>c:{c}', f'<td style="background-color:{c}">') for c in colors.values()]
    for fs, ns in replacements:
        html_str = html_str.replace(fs, ns)
    return HTMLDisplay(html_str)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hela',
 'hela._utils',
 'hela.datasets',
 'hela.math',
 'hela.plots',
 'hela.test_suite',
 'hela.web_page']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.20.1,<2.0.0', 'pandas>=1.2.3,<2.0.0']

extras_require = \
{'bigquery': ['google-cloud-bigquery>=2,<3'],
 'glue': ['aws-cdk.aws-glue>=1,<2'],
 'spark': ['pyspark>=3,<4']}

setup_kwargs = {
    'name': 'hela',
    'version': '0.2.1',
    'description': 'Your data catalog as code and one schema to rule them all.',
    'long_description': '# `hela`: write your data catalog as code\n![Unit Tests](https://github.com/erikmunkby/hela/actions/workflows/unit_tests.yaml/badge.svg)\n![Spark](https://github.com/erikmunkby/hela/actions/workflows/test_spark.yaml/badge.svg)\n![BigQuery](https://github.com/erikmunkby/hela/actions/workflows/test_bigquery.yaml/badge.svg)\n![AWS Glue](https://github.com/erikmunkby/hela/actions/workflows/test_aws_glue.yaml/badge.svg)\n\nYou probably already have your data job scripts version controlled, but what about your data catalog?\nThe answer: **write your data catalog as code!** Storing your data catalog and data documentation as code makes your catalog searchable, referenceable, reliable, platform agnostic, sets you up for easy collaboration and much more! \nThis library is built to fit small and large data landscapes, but is happiest when included from the start.\n\n`Hela` (or Hel) is the norse mythological collector of souls, and the Swedish word for "whole" or "all of it". `Hela`\nis designed to give everyone a chance to build a data catalog, with a low entry barrier: pure python code.\n\nTODO: Add link to homepage, pypi?, demo repo\n\n## Installing\nUsing pip:\n\n`pip install hela`\n\nUsing poetry:\n\n`poetry add hela`\n\n## Roadmap\nThese are up-coming features in no particular order, but contributions towards these milestones are highly appreciated! To read more about contributing check out `CONTRIBUTING.md`.\n\n* Search functionality in web app\n* More integrations (Snowflake, Redshift)\n* More feature rich dataset classes\n* Data lineage functionality (both visualized in notebooks and web app)\n* Prettier docs page\n\n\n## (Mega) Quick start\nIf you want to read more check out the [docs page](TODO: insert docs page link). If you do not have patience for that, the following is all you need to get started.\n\nFirst of all build your own dataset class by inheriting the `BaseDataset` class. This class will hold most of your project specific functionality such as read/write, authentication etc.\n\n```python\nfrom hela import BaseDataset, Col\nfrom hela.data_types import String\n\nclass MyDatasetClass(BaseDataset):\n    def __init__(\n        self,\n        name: str, # Required\n        description: str, # Optional but recommended\n        columns: list, # Optional but recommended\n        rich_description_path: str = None, # Optional, used for web app\n        partition_cols: list = None,  # Optional but recommended\n        # folder: str = None, # Only do one of either folder or database\n        database: str, # Optional, can also be enriched via Catalog\n    ) -> None:\n        super().__init__(\n            name,\n            data_type=\'bigquery\',\n            folder=None,\n            database=database,\n            description=description,\n            rich_description_path=rich_description_path,\n            partition_cols=partition_cols,\n            dependencies=None,\n            columns=columns\n        )\n        # Do more of your own init stuff\n\n    def my_func(self) -> None:\n        # Your own dataset function\n        pass\n        \n# Now instantiate your dataset class with one example column\nmy_dataset = MyDatasetClass(\'my_dataset\', \'An example dataset.\', [\n    Col(\'my_column\', String(), \'An example column.\')\n])\n```\n\nNow that you have a dataset class, and instantiated your first dataset, you can start populating your\ndata catalog.\n\n```python\nfrom hela import Catalog\n\nclass MyCatalog(Catalog):\n    my_dataset = my_dataset\n```\n\nThat\'s it! You now have a small catalog to keep building on. To view it as a web page you can\nadd the following code to a python script, and in the future add it in whichever CI/CD tool you use.\nThis will generate an `index.html` file that you can view in your browser or host on e.g. github pages.\n\n```python\nfrom hela import generate_webpage\n\ngenerate_webpage(MyCatalog, output_folder=\'.\')\n```\n\nTo view what a bigger data catalog can look like check out the [showcase catalog](TODO: add link to showcase repo).',
    'author': 'Erik Munkby',
    'author_email': 'erik.munkby@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/erikmunkby/hela',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)



from __future__ import annotations
import warnings
import pandas as pd
import uuid
from pathlib import Path
from abc import ABC
from typing import Optional, Set, Union, Sequence, Dict, Any, Tuple
from datetime import date
from collections import Counter
from dataclasses import dataclass
from hela._utils.date_utils import get_missing_dates
from hela._utils.dict_utils import flatten_dict
from hela._column_classes import _ColumnType
from hela._constants import _PATH_VAR
from hela._utils.path_utils import join_paths
from hela.errors import DatasetError, DuplicationError
from hela.plots.date_plots import plot_date_availability_calendar
from hela._utils.short_description import ShortDescription


@dataclass
class _DatasetInfo:
    name: str
    data_type: str
    description: str
    folder: str = None
    min_date: date = None
    max_date: date = None
    nbr_missing_dates: int = None


class Columns(Tuple[_ColumnType]):
    pass


class BaseDataset(ABC):
    """Abstract Dataset class to be used when defining building your own datasets.

    If you choose to build data interactivity through the data catalog, it is within
    your own dataset classes you would build authentication and connection logic.

    For full usage of the available catalog features implement the functions
    `BaseDataset.get_samples` and `BaseDataset.get_dates`.

    Attributes:
        name: The name of the dataset
        data_type: The data type of the dataset e.g. "parquet" or "bigquery
        description: A description of the dataset as a string
        partition_cols: A list of column names to be used for partitioning as strings
        rich_description_path:  A path to a markdown file with possibilities for longer,
                                more detailed descriptions. Primarily used for generated catalog web page.
        columns: A list of class ColumnType objects defining the columns of the dataset
        path: The path to the dataset (combination of folder and name)
    """
    _is_dataset: bool = True
    _type: str = 'Dataset'

    def __init__(
        self,
        name: str,
        data_type: str,
        folder: Optional[Union[str, Path]] = None,
        database: Optional[str] = None,
        description: Optional[str] = None,
        rich_description_path: Optional[str] = None,
        partition_cols: Optional[Sequence[str]] = None,
        columns: Optional[Sequence[_ColumnType]] = None,
    ) -> None:
        self.name = name
        self.data_type = data_type
        self.description = description
        self.rich_description_path = rich_description_path
        self.partition_cols = partition_cols
        self.database = database
        self.folder = folder
        self.path = None
        self._set_path()
        self._set_columns(columns)
        # _id used to build links in generated catalog website
        self._id: str = str(uuid.uuid4())

    def _set_columns(self, columns: Optional[Sequence[_ColumnType]] = None) -> None:
        if columns is None:
            self.columns = columns
            return
        duplicated_columns = ', '.join(
            [f'"{col.name}"' for col, count in Counter(columns).items() if count > 1]
        )
        if duplicated_columns:
            raise DuplicationError(f'Found duplication of column(s) {duplicated_columns} in dataset "{self.name}".')
        col_list = Columns(columns)
        for c in columns:
            setattr(col_list, c.name, c)
        self.columns = col_list

    def _set_path(self) -> None:
        if self.folder is None:
            return
        path = join_paths(self.folder, self.name).with_suffix(f'.{self.data_type}')
        self.path = path
        setattr(self, _PATH_VAR, path)

    def _describe(self) -> _DatasetInfo:
        info_obj = _DatasetInfo(
            name=self.name,
            data_type=self.data_type,
            description=self.description
        )
        try:
            dates = self.get_dates()
            if dates is None:
                return info_obj
            info_obj.min_date = min(dates)
            info_obj.max_date = max(dates)
            info_obj.nbr_missing_dates = len(get_missing_dates(dates))
        except NotImplementedError:
            pass
        return info_obj

    def show_columns(self, samples: bool = True) -> pd.DataFrame:
        """Returns a dataframe with information of the columns of this dataset, one column per row.

        Args:
            samples:    When true will include a sample datapoint for all columns.
                        Requires implementation of `BaseDataset.get_samples` function.

        Returns:
            A pandas dataframe with one column per row.
        """
        if self.columns is None:
            return None

        column_df = pd.DataFrame([
            cinfo.__dict__ for c in self.columns for cinfo in c._describe()
        ])
        if samples:
            try:
                fetched_samples = self.get_samples()
                if fetched_samples:
                    fetched_samples = {**fetched_samples, **flatten_dict(fetched_samples)}
                    column_df.loc[:, 'Sample'] = column_df.name.apply(lambda x: fetched_samples.get(x, None))
            except NotImplementedError:
                pass
        return column_df

    def show_dates(self) -> None:
        """
        Will generate a grid plot of all available dates for this dataset.
        Requires `BaseDataset.get_dates` implemented.
        """
        dates = self.get_dates()
        if dates is None:
            raise ValueError(f'No dates could be fetched from dataset {self}')
        return plot_date_availability_calendar(dates)

    def check_columns(
        self,
        column_list: Sequence[str],
        raise_undefined_columns: bool = False,
        raise_missing_columns: bool = False
    ) -> None:
        """
        Will compare the sent in column list against the dataset's defined columns
        and inform (warn or raise) regarding any discrepancies.

        Args:
            column_list: A list of names of columns as strings
            raise_undefined_columns: Optional; If True will raise if columns
                found in column_list not defined in dataset
            raise_missing_columns: If True will raise if columns
                defined in dataset not found in column_list

        Raises:
            DatasetError: If any of raise flags are set to True

        Examples:
        >>> my_dataset.check_columns(df.columns, raise_undefined_columns=True)
        """
        if self.columns is None:
            warnings.warn('Dataset has no columns specified.')
            return
        undefined_columns = set(column_list) - set([c.name for c in self.columns])
        msg = f'The following columns are not defined in dataset: {list(undefined_columns)}'
        if raise_undefined_columns:
            raise DatasetError(msg)
        warnings.warn(msg)

        missing_columns = set([c.name for c in self.columns]) - set(column_list)
        msg = f'The following columns are missing from column_list: {list(missing_columns)}'
        if raise_missing_columns:
            raise DatasetError(msg)
        warnings.warn(msg)

    @property
    def _prefix(self) -> str:
        """Returns the prefix as either folder, database or empty string."""
        if self.folder:
            return self.folder
        if self.database:
            return self.database
        return ''

    def __str__(self) -> str:
        prefix = self._prefix
        if prefix:
            return f'{prefix}:{self.name}'
        return self.name

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, o: BaseDataset) -> bool:
        return self.name == o.name and self._id == o._id

    def __hash__(self) -> int:
        return hash(self.__str__())

    def _desc_(self) -> ShortDescription:
        return ShortDescription(name=self.name, type=self._type, description=self.description)

    def get_dates(self) -> Optional[Set[date]]:
        """Implement this function for date inspection functionality such as `BaseDataset.show_dates`.

        Should return a set of dates when called or None if dates for some reason could not be fetched.
        """
        raise NotImplementedError

    def get_samples(self) -> Optional[Dict[str, Any]]:
        """Implement this function for sample inspection functionality used in e.g. `BaseDataset.show_columns`.

        Should return a dictionary of string keys for column names with samples:
        >>> {'my_column': 123}

        Nested columns should return names with dot-notation:
        >>> {'parent_column.my_column': 123}

        Or None if samples could not be fetched:
        >>> None
        """
        raise NotImplementedError



import numpy as np
from typing import Sequence
from hela.math.preprocessing import clean_corpus, clean_document
from hela.math.score_info import SimilarityInfo


def distance(s1: str, s2: str) -> float:
    size_x, size_y = len(s1) + 1, len(s2) + 1

    matrix = np.zeros((size_x, size_y), dtype=int)

    # Populate the top row, and leftmost column in the matrix
    for x in range(1, size_x):
        matrix[x, 0] = x
    for y in range(1, size_y):
        matrix[0, y] = y

    for x in range(1, size_x):
        for y in range(1, size_y):
            # If the characters equal eachother we have 0 cost
            cost = 0 if s1[x - 1] == s2[y - 1] else 1

            matrix[x, y] = min(
                matrix[x - 1, y] + 1,  # Deletions
                matrix[x, y - 1] + 1,  # Insertions
                matrix[x - 1, y - 1] + cost  # Substitutions
            )
    return round(1 - matrix[-1, -1] / max(len(s1), len(s2)), 3)


def sort(query_str: str, corpus: Sequence[str], min_similarity: float = .75) -> Sequence[SimilarityInfo]:
    """Sorts a corpus of documents based on levenshtein distance similarity.

    Args:
        query_str:      The string to compare against the corpus
        corpus:         A sequence of documents as strings
        min_similarity: The minimum required similarity [0 to 1]

    Returns:
        A list of SimilarityInfo Objects
    """
    documents, tokenized_documents = clean_corpus(corpus)
    query_str = clean_document(query_str)

    matches = [
        SimilarityInfo(
            score=distance(query_str, document),
            match_idx=i,
            match_string=documents[i],
            target_string=query_str
        )
        for i, document in enumerate(tokenized_documents)
    ]
    matches = [m for m in matches if m.score >= min_similarity]
    return sorted(matches, key=lambda x: -x.score)


def replace(query_str: str, corpus: Sequence[str], min_similarity: float = .75) -> str:
    # Create boundaries for length of replacement string
    boundary_length = round(len(query_str) * (1 - min_similarity))
    min_len, max_len = len(query_str) - boundary_length, len(query_str) + boundary_length
    # Slim down corpus based on string length
    slimmed_corpus = [doc for doc in corpus if len(doc) >= min_len and len(doc) <= max_len]
    results = sort(query_str, slimmed_corpus, min_similarity=min_similarity)
    if len(results) > 0:
        return results[0].match_string
    return query_str
